//@+ TerminationPolicy = never

/*@ predicate edge{M}(Cell x, Cell y) =
      x != null && (x.left == y || x.right == y) ; */

/*@ inductive path{M}(Cell x, Cell y) {
    case path_nil{M}:
      \forall Cell x;
      path(x, x);
    case path_cons{M}:
      \forall Cell x, Cell y, Cell z;
      path(x, y) && edge(y, z) ==> path(x, z);
    } */

/*@ predicate only_descendants_are_marked{M} (Cell root) =
      \forall Cell x;
      x != null && x.marked == true ==>
      path(root, x) ; */

/*@ predicate well_colored{M} () =
      \forall Cell x, Cell y;
      edge(x, y) && y != null ==>
      x.busy == true || (x.marked == true ==> y.marked == true) ; */

/*@ predicate all_descendants_are_marked{M} (Cell root) =
      root != null ==>
      root.marked == true &&
      \forall Cell x, Cell y;
      edge(x, y) && y != null ==>
      x.marked == true ==> y.marked == true ; */

/*@ lemma reformulation{M}:
      \forall Cell root;
      all_descendants_are_marked(root) ==>
      \forall Cell x;
      x != null && path(root, x) ==>
      x.marked == true ; */

class Cell {

  // A cell has two fields.
  // Each field holds a (possibly null) pointer to a cell.
  // Thus, cells form a (possibly cyclic) graph.

  Cell left, right;

  // A cell carries a Boolean mark.

  boolean marked;

  // A cell carries a Boolean flag that indicates that a call to dfs for
  // this cell is in progress. A busy cell can be thought of as ``grey''.

  //@ ghost boolean busy;

  // This global variable is supposed to hold the root of the graph.

  static Cell root;

  // Depth-first search.

  /*@ requires well_colored()
	    && only_descendants_are_marked(root)
	    && path(root, c)
	     ;
       ensures well_colored()
	    && only_descendants_are_marked(root)
	    && (\forall Cell x; x.busy == true ==> \old(x.busy) == true)     // fewer nodes are busy
	    && (\forall Cell x; \old(x.marked) == true ==> x.marked == true) // more nodes are marked
	    && (c != null ==> c.marked == true)                              // c is marked
	     ; */
  static void dfs (Cell c)
  {
    if (c != null && !c.marked) {
      //@ ghost c.busy = true;
      c.marked = true;
      dfs(c.left);
      dfs(c.right);
      //@ ghost c.busy = false;
    }
  }

  // The root procedure.

  /*@ requires (\forall Cell x; x != null ==> x.marked == false && x.busy == false)
	     ;
       ensures all_descendants_are_marked(root)
	    && only_descendants_are_marked(root)
	    && (\forall Cell x; x != null ==> x.busy == false)
	     ; */
  static void traverse ()
  {
    //@ assert well_colored() ;
    dfs(root);
  }

}

