//@+ CheckArithOverflow = no

// If n is odd, then n % 4 is 1 or 3.
/*@
lemma basic0:
  \forall integer n;
  n > 0 && n % 2 == 1 ==>
  n % 4 == 1 || n % 4 == 3
;
*/

// If n is odd, then n % 8 is 1 or 3 or 5 or 7.
/*@
lemma basic1:
  \forall integer n;
  n > 0 && n % 2 == 1 ==>
  n % 8 == 1 || n % 8 == 3 || n % 8 == 5 || n % 8 == 7
;
*/

// (_ % n) is idempotent.
/*@
lemma basic2:
  \forall integer a n;
  (a % n) % n == a % n
;
*/

/*@ axiomatic Jacobi {

// We axiomatize jacobi(a, n) as a total function, even though it is
// defined only when n is positive and odd. We add this condition
// about n in the axioms below, when necessary.

logic integer jacobi(integer a, integer n);

axiom jacobi1:
  \forall integer n;
  n > 0 && n % 2 == 1 ==>
  jacobi(0, n) == 0
;

axiom jacobi2:
  \forall integer n;
  n > 0 && n % 2 == 1 ==>
  jacobi(1, n) == 1
;

axiom jacobi3a:
  \forall integer n;
  n > 0 && n % 2 == 1 ==>
  n % 8 == 1 || n % 8 == 7 ==>
  jacobi(2, n) == 1
;

axiom jacobi3b:
  \forall integer n;
  n > 0 && n % 2 == 1 ==>
  n % 8 == 3 || n % 8 == 5 ==>
  jacobi(2, n) == -1
;

axiom jacobi4:
  \forall integer a b n;
  n > 0 && n % 2 == 1 ==>
  a % n == b % n ==>
  jacobi(a, n) == jacobi(b, n)
;

axiom jacobi5:
  \forall integer a b n;
  n > 0 && n % 2 == 1 ==>
  jacobi(a * b, n) == jacobi(a, n) * jacobi(b, n)
;

// The following axioms are the quadratic reciprocity laws.
// Wikipedia says they are true when a and n are coprime,
// i.e., when gcd(a, n) == 1. However, if a and n are not
// coprime, then jacobi(a, n) == jacobi(n, a) == 0, so the
// laws still hold.

axiom jacobi6a:
  \forall integer a n;
  a > 0 && a % 2 == 1 ==>
  n > 0 && n % 2 == 1 ==>
  a % 4 == 1 || n % 4 == 1 ==>
  jacobi(a, n) == jacobi(n, a)
;

axiom jacobi6b:
  \forall integer a n;
  a > 0 && a % 2 == 1 ==>
  n > 0 && n % 2 == 1 ==>
  a % 4 == 3 && n % 4 == 3 ==>
  jacobi(a, n) == - jacobi(n, a)
;

} */

class Jacobi {

  /*@ requires 0 <= a && n > 0 && n % 2 == 1;
      decreases a;
      ensures  \result == jacobi(a, n) ; */
  static int jacobiRec (int a, int n)
  {
    if (a == 0)
      return 0;
    if (a == 1)
      return 1;
    if (a == 2)
      if (n % 8 == 1 || n % 8 == 7)
        return 1;
      else
        return -1;
    if (a % 2 == 0)
      return jacobiRec (2, n) * jacobiRec (a / 2, n);
    if (a % 4 == 1 || n % 4 == 1)
      return jacobiRec (n % a, a);
    else
      return -jacobiRec (n % a, a);
  }

  /*@ requires 0 <= ia && in > 0 && in % 2 == 1;
      ensures  \result == jacobi(ia, in) ; */
  static int jacobiLoop (final int ia, final int in)
  {
    int a = ia;
    int n = in;
    int sign = 1;
    /*@ loop_invariant
          0 <= a && n > 0 && n % 2 == 1 &&
          jacobi(ia, in) == sign * jacobi(a, n);
        loop_variant
          a;
    */
    while (a > 2) {
      if (a % 2 == 0) {
        if (n % 8 == 3 || n % 8 == 5)
          sign = -sign;
        a /= 2;
      }
      else {
        if (a % 4 == 3 && n % 4 == 3)
          sign = -sign;
        int tmp = a;
        a = n % a;
        n = tmp;
      }
    }
    if (a == 0)
      return 0;
    if (a == 1)
      return sign;
    if (a == 2)
      if (n % 8 == 3 || n % 8 == 5)
        return -sign;
      else
        return sign;
    // unreachable
    return 0;
  }

}

